@@ -1,6 +1,17 @@
 import logging
+import sys
+
+PY3 = sys.version_info[0] == 3
+PY2 = not PY3
+
+__all__ = ['BotPlugin', 'webhook', 'webroute', 'webview']
 
 from errbot.botplugin import BotPlugin  # repeat it here for convenience and coherence with @botcmd
+from errbot.builtins.wsview import webhook, route, view  # idem repeat here the webhook feature for convenience to expose webhooks
+
+
+webroute = route  # this allows plugins to expose dynamic webpages on err embedded webserver
+webview = view  # this allows to use the templating system
 
 
 def botcmd(*args, **kwargs):
