@@ -1,11 +1,11 @@
-from collections import deque
+from collections import deque, defaultdict
 import inspect
 import logging
-from pyexpat import ExpatError
-from xmpp.simplexml import XML2Node
-from errbot import botcmd
+from xml.etree import cElementTree as ET
+from xml.etree.cElementTree import ParseError
+from errbot import botcmd, PY2
 import difflib
-from errbot.utils import get_sender_username, xhtml2txt, get_jid_from_message, utf8, parse_jid
+from errbot.utils import get_sender_username, xhtml2txt, utf8, parse_jid, split_string_after
 from errbot.templating import tenv
 import traceback
 
@@ -21,6 +21,16 @@ try:
 except ImportError:
     ACCESS_CONTROLS = {}
 
+try:
+    from config import HIDE_RESTRICTED_COMMANDS
+except ImportError:
+    HIDE_RESTRICTED_COMMANDS = False
+
+try:
+    from config import HIDE_RESTRICTED_ACCESS
+except ImportError:
+    HIDE_RESTRICTED_ACCESS = False
+
 try:
     from config import BOT_PREFIX_OPTIONAL_ON_CHAT
 except ImportError:
@@ -48,6 +58,11 @@ except ImportError:
     logging.warning("DIVERT_TO_PRIVATE is missing in config")
     pass
 
+try:
+    from config import MESSAGE_SIZE_LIMIT
+except ImportError:
+    MESSAGE_SIZE_LIMIT = 10000  # Corresponds with what HipChat accepts
+
 if BOT_ASYNC:
     from errbot.bundled.threadpool import ThreadPool, WorkRequest
 
@@ -89,7 +104,7 @@ class Identifier(object):
         return answer
 
     def __unicode__(self):
-        return unicode(self.__str__())
+        return str(self.__str__())
 
 
 class Message(object):
@@ -97,12 +112,14 @@ class Message(object):
 
     def __init__(self, body, typ='chat', html=None):
         # it is either unicode or assume it is utf-8
-        if isinstance(body, unicode):
+        if isinstance(body, str):
             self.body = body
         else:
             self.body = body.decode('utf-8')
         self.html = html
         self.typ = typ
+        self.delayed = False
+        self.mucknick = None
 
     def setTo(self, to):
         if isinstance(to, Identifier):
@@ -128,33 +145,26 @@ class Message(object):
         else:
             self.fr = Identifier(fr)  # assume a parseable string
 
-    def getProperties(self):
-        return {}
-
     def getBody(self):
         return self.body
 
     def getHTML(self):
         return self.html
 
-    # XMPP backward compliance
-    def getTagAttr(self, tag, attr):
-        return None
+    def setHTML(self, html):
+        self.html = html
 
-    def getTagData(self, tag):
-        return None
+    def setDelayed(self, delayed):
+        self.delayed = delayed
 
-    def getTag(self, tag):
-        return None
+    def isDelayed(self):
+        return self.delayed
 
-    def addChild(self, name=None, attrs={}, payload=[], namespace=None, node=None):
-        """ If "node" argument is provided, adds it as child node. Else creates new node from
-            the other arguments' values and adds it as well."""
-        if node.name == 'html':
-            self.html = unicode(node)  # assume this is the html node you want to set
-        else:
-            raise TypeError('We only support the custom html node from XMPP')
-        return node
+    def setMuckNick(self, nick):
+        self.mucknick = nick
+
+    def getMuckNick(self):
+        return self.mucknick
 
     def __str__(self):
         return self.body
@@ -170,32 +180,34 @@ def build_text_html_message_pair(source):
     text_plain = None
 
     try:
-        node = XML2Node(source)
+        node = ET.XML(source)
         text_plain = xhtml2txt(source)
-    except ExpatError as ee:
+    except ParseError as ee:
         if source.strip():  # avoids keep alive pollution
             logging.debug('Could not parse [%s] as XHTML-IM, assume pure text Parsing error = [%s]' % (source, ee))
             text_plain = source
     return text_plain, node
 
+
 def build_message(text, message_class, conversion_function=None):
     """Builds an xhtml message without attributes.
     If input is not valid xhtml-im fallback to normal."""
     message = None  # keeps the compiler happy
     try:
-        text = utf8(text)
-
         text = text.replace('', '*')  # there is a weird chr IRC is sending that we need to filter out
+        if PY2:
+            ET.XML(text.encode('utf-8'))  # test if is it xml
+        else:
+            ET.XML(text)
 
-        XML2Node(text)  # test if is it xml
         edulcorated_html = conversion_function(text) if conversion_function else text
         try:
             text_plain, node = build_text_html_message_pair(edulcorated_html)
             message = message_class(body=text_plain)
-            message.addChild(node=node)
-        except ExpatError as ee:
+            message.setHTML(node)
+        except ET.ParseError as ee:
             logging.error('Error translating to hipchat [%s] Parsing error = [%s]' % (edulcorated_html, ee))
-    except ExpatError as ee:
+    except ET.ParseError as ee:
         if text.strip():  # avoids keep alive pollution
             logging.debug('Determined that [%s] is not XHTML-IM (%s)' % (text, ee))
         message = message_class(body=text)
@@ -205,18 +217,17 @@ def build_message(text, message_class, conversion_function=None):
 class Backend(object):
     # Implements the basic Bot logic (logic independent from the backend) and leave to you to implement the missing parts
 
-    cmd_history = deque(maxlen=10)
+    cmd_history = defaultdict(lambda: deque(maxlen=10)) #this will be a per user history
+
     MSG_ERROR_OCCURRED = 'Sorry for your inconvenience. '\
                          'An unexpected error occurred.'
-    MESSAGE_SIZE_LIMIT = 10000 # the default one from hipchat
-    MESSAGE_SIZE_ERROR_MESSAGE = '|<- SNIP ! Message too long.'
+    MESSAGE_SIZE_LIMIT = MESSAGE_SIZE_LIMIT
     MSG_UNKNOWN_COMMAND = 'Unknown command: "%(command)s". '\
                           'Type "' + BOT_PREFIX + 'help" for available commands.'
     MSG_HELP_TAIL = 'Type help <command name> to get more info '\
                     'about that specific command.'
     MSG_HELP_UNDEFINED_COMMAND = 'That command is not defined.'
 
-
     def __init__(self, *args, **kwargs):
         """ Those arguments will be directly those put in BOT_IDENTITY
         """
@@ -230,7 +241,6 @@ class Backend(object):
         else:
             self.bot_alt_prefixes = BOT_ALT_PREFIXES
 
-
     def send_message(self, mess):
         """Send a message"""
         self.connect().send_message(mess)
@@ -244,9 +254,6 @@ class Backend(object):
         Message is NOT sent"""
         response = self.build_message(text)
         if private:
-            # Use get_jid_from_message here instead of mess.getFrom because
-            # getFrom will return the groupchat id instead of user's jid when
-            # sent from a chatroom
             response.setTo(get_jid_from_message(mess))
             response.setType('chat')
             response.setFrom(self.jid)
@@ -263,15 +270,18 @@ class Backend(object):
         # Prepare to handle either private chats or group chats
         type = mess.getType()
         jid = mess.getFrom()
-        props = mess.getProperties()
         text = mess.getBody()
         username = get_sender_username(mess)
+        user_cmd_history = self.cmd_history[username];
+
+        if mess.isDelayed():
+            logging.debug("Message from history, ignore it")
+            return False
 
         if type not in ("groupchat", "chat"):
             logging.debug("unhandled message type %s" % mess)
             return False
 
-        logging.debug("*** props = %s" % props)
         logging.debug("*** jid = %s" % jid)
         logging.debug("*** username = %s" % username)
         logging.debug("*** type = %s" % type)
@@ -279,7 +289,8 @@ class Backend(object):
 
         # If a message format is not supported (eg. encrypted),
         # txt will be None
-        if not text: return False
+        if not text:
+            return False
 
         surpress_cmd_not_found = False
 
@@ -311,7 +322,7 @@ class Backend(object):
             surpress_cmd_not_found = True
         elif not text.startswith(BOT_PREFIX):
             return True
-        else:
+        if text.startswith(BOT_PREFIX):
             text = text[len(BOT_PREFIX):]
 
         text_split = text.strip().split(' ')
@@ -333,96 +344,51 @@ class Backend(object):
                     args = ' '.join(text_split[1:])
 
         if command == BOT_PREFIX:  # we did "!!" so recall the last command
-            if len(self.cmd_history):
-                cmd, args = self.cmd_history[-1]
+            if len(user_cmd_history):
+                cmd, args = user_cmd_history[-1]
             else:
                 return False  # no command in history
         elif command.isdigit():  # we did "!#" so we recall the specified command
             index = int(command)
-            if len(self.cmd_history) >= index:
-                cmd, args = self.cmd_history[-index]
+            if len(user_cmd_history) >= index:
+                cmd, args = user_cmd_history[-index]
             else:
                 return False  # no command in history
 
-        if (cmd, args) in self.cmd_history:
-            self.cmd_history.remove((cmd, args))  # we readd it below
-
-        logging.info("received command = %s matching [%s] with parameters [%s]" % (command, cmd, args))
+        if (cmd, args) in user_cmd_history:
+            user_cmd_history.remove((cmd, args))  # we readd it below
 
         if cmd:
-            def execute_and_send(template_name):
-                try:
-                    reply = self.commands[cmd](mess, args)
+            logging.info("received command = %s matching [%s] with parameters [%s]" % (command, cmd, args))
 
-                    # integrated templating
-                    if template_name:
-                        reply = tenv().get_template(template_name + '.html').render(**reply)
-
-                    # Reply should be all text at this point (See https://github.com/gbin/err/issues/96)
-                    reply = unicode(reply)
-                except Exception, e:
-                    logging.exception(u'An error happened while processing '
-                                      u'a message ("%s") from %s: %s"' %
-                                      (text, jid, traceback.format_exc(e)))
-                    reply = self.MSG_ERROR_OCCURRED + ':\n %s' % e
-                if reply:
-                    if len(reply) > self.MESSAGE_SIZE_LIMIT:
-                        reply = reply[:self.MESSAGE_SIZE_LIMIT - len(self.MESSAGE_SIZE_ERROR_MESSAGE)] + self.MESSAGE_SIZE_ERROR_MESSAGE
-                    self.send_simple_reply(mess, reply, cmd in DIVERT_TO_PRIVATE)
-
-            usr = str(get_jid_from_message(mess))
-            typ = mess.getType()
-            if cmd not in ACCESS_CONTROLS:
-                ACCESS_CONTROLS[cmd] = ACCESS_CONTROLS_DEFAULT
-
-            if 'allowusers' in ACCESS_CONTROLS[cmd] and usr not in ACCESS_CONTROLS[cmd]['allowusers']:
-                self.send_simple_reply(mess, "You're not allowed to access this command from this user")
+            access, accessError = self.check_command_access(mess, cmd)
+            if not access:
+                if not HIDE_RESTRICTED_ACCESS:
+                    self.send_simple_reply(mess, accessError)
                 return False
-            if 'denyusers' in ACCESS_CONTROLS[cmd] and usr in ACCESS_CONTROLS[cmd]['denyusers']:
-                self.send_simple_reply(mess, "You're not allowed to access this command from this user")
-                return False
-            if typ == 'groupchat':
-                stripped = mess.getFrom().getStripped()
-                if 'allowmuc' in ACCESS_CONTROLS[cmd] and ACCESS_CONTROLS[cmd]['allowmuc'] is False:
-                    self.send_simple_reply(mess, "You're not allowed to access this command from a chatroom")
-                    return False
-                if 'allowrooms' in ACCESS_CONTROLS[cmd] and stripped not in ACCESS_CONTROLS[cmd]['allowrooms']:
-                    self.send_simple_reply(mess, "You're not allowed to access this command from this room")
-                    return False
-                if 'denyrooms' in ACCESS_CONTROLS[cmd] and stripped in ACCESS_CONTROLS[cmd]['denyrooms']:
-                    self.send_simple_reply(mess, "You're not allowed to access this command from this room")
-                    return False
-            else:
-                if 'allowprivate' in ACCESS_CONTROLS[cmd] and ACCESS_CONTROLS[cmd]['allowprivate'] is False:
-                    self.send_simple_reply(mess, "You're not allowed to access this command via private message to me")
-                    return False
 
             f = self.commands[cmd]
 
-            if f._err_command_admin_only:
-                if typ == 'groupchat':
-                    self.send_simple_reply(mess, 'You cannot administer the bot from a chatroom, message the bot directly')
-                    return False
-                if usr not in BOT_ADMINS:
-                    self.send_simple_reply(mess, 'You cannot administer the bot from this user %s.' % usr)
-                    return False
-                if BOT_ASYNC:
-                    self.thread_pool.wait() # If it is an admin command, wait that the queue is completely depleted so we don't have strange concurrency issues on load/unload/updates etc ...
+            if f._err_command_admin_only and BOT_ASYNC:
+                    self.thread_pool.wait()  # If it is an admin command, wait that the queue is completely depleted so we don't have strange concurrency issues on load/unload/updates etc ...
 
             if f._err_command_historize:
-                self.cmd_history.append((cmd, args)) # add it to the history only if it is authorized to be so
+                user_cmd_history.append((cmd, args))  # add it to the history only if it is authorized to be so
 
             # Don't check for None here as None can be a valid argument to split.
             # '' was chosen as default argument because this isn't a valid argument to split()
             if f._err_command_split_args_with != '':
                 args = args.split(f._err_command_split_args_with)
             if BOT_ASYNC:
-                wr = WorkRequest(execute_and_send, [f._err_command_template]) #execute_and_send(f._err_command_template)
+                wr = WorkRequest(self._execute_and_send,
+                                 [], {'cmd': cmd, 'args': args, 'mess': mess, 'jid': jid,
+                                      'template_name': f._err_command_template})
                 self.thread_pool.putRequest(wr)
                 if f._err_command_admin_only:
-                    self.thread_pool.wait() # Again wait for the completion before accepting a new command that could generate weird concurrency issues
+                    self.thread_pool.wait()  # Again wait for the completion before accepting a new command that could generate weird concurrency issues
             else:
-                execute_and_send(f._err_command_template)
+                self._execute_and_send(cmd=cmd, args=args, mess=mess, jid=jid,
+                                       template_name=f._err_command_template)
 
         else:
             logging.debug("Command not found")
@@ -437,6 +403,78 @@ class Backend(object):
 
         return True
 
+    def _execute_and_send(self, cmd, args, mess, jid, template_name=None):
+        """Execute a bot command and send output back to the caller
+
+        cmd: The command that was given to the bot (after being expanded)
+        args: Arguments given along with cmd
+        mess: The message object
+        jid: The jid of the person executing the command
+        template_name: The name of the template which should be used to render
+            html-im output, if any
+
+        """
+
+        def process_reply(reply):
+            # integrated templating
+            if template_name:
+                reply = tenv().get_template(template_name + '.html').render(**reply)
+
+            # Reply should be all text at this point (See https://github.com/gbin/err/issues/96)
+            return str(reply)
+
+        def send_reply(reply):
+            for part in split_string_after(reply, self.MESSAGE_SIZE_LIMIT):
+                self.send_simple_reply(mess, part, cmd in DIVERT_TO_PRIVATE)
+
+        try:
+            if inspect.isgeneratorfunction(self.commands[cmd]):
+                replies = self.commands[cmd](mess, args)
+                for reply in replies:
+                    if reply: send_reply(process_reply(reply))
+            else:
+                reply = self.commands[cmd](mess, args)
+                if reply: send_reply(process_reply(reply))
+        except Exception as e:
+            tb = traceback.format_exc()
+            logging.exception('An error happened while processing '
+                              'a message ("%s") from %s: %s"' %
+                              (mess.getBody(), jid, tb))
+            send_reply(self.MSG_ERROR_OCCURRED + ':\n %s' % e)
+
+    def check_command_access(self, mess, cmd):
+        usr = str(get_jid_from_message(mess))
+        typ = mess.getType()
+
+        if cmd not in ACCESS_CONTROLS:
+            ACCESS_CONTROLS[cmd] = ACCESS_CONTROLS_DEFAULT
+
+        if 'allowusers' in ACCESS_CONTROLS[cmd] and usr not in ACCESS_CONTROLS[cmd]['allowusers']:
+            return False, "You're not allowed to access this command from this user"
+        if 'denyusers' in ACCESS_CONTROLS[cmd] and usr in ACCESS_CONTROLS[cmd]['denyusers']:
+            return False, "You're not allowed to access this command from this user"
+        if typ == 'groupchat':
+            stripped = mess.getFrom().getStripped()
+            if 'allowmuc' in ACCESS_CONTROLS[cmd] and ACCESS_CONTROLS[cmd]['allowmuc'] is False:
+                return False, "You're not allowed to access this command from a chatroom"
+            if 'allowrooms' in ACCESS_CONTROLS[cmd] and stripped not in ACCESS_CONTROLS[cmd]['allowrooms']:
+                return False, "You're not allowed to access this command from this room"
+            if 'denyrooms' in ACCESS_CONTROLS[cmd] and stripped in ACCESS_CONTROLS[cmd]['denyrooms']:
+                return False, "You're not allowed to access this command from this room"
+        else:
+            if 'allowprivate' in ACCESS_CONTROLS[cmd] and ACCESS_CONTROLS[cmd]['allowprivate'] is False:
+                return False, "You're not allowed to access this command via private message to me"
+
+        f = self.commands[cmd]
+
+        if f._err_command_admin_only:
+            if typ == 'groupchat':
+                return False, 'You cannot administer the bot from a chatroom, message the bot directly'
+            if usr not in BOT_ADMINS:
+                return False, 'You cannot administer the bot from this user %s.' % usr
+        
+        return True, ""
+
     def unknown_command(self, mess, cmd, args):
         """ Override the default unknown command behavior
         """
@@ -464,7 +502,7 @@ class Backend(object):
                 if name in self.commands:
                     f = self.commands[name]
                     new_name = (classname + '-' + name).lower()
-                    self.warn_admins('%s.%s clashes with %s.%s so it has been renamed %s' % (classname, name, f.im_class.__name__, f.__name__, new_name ))
+                    self.warn_admins('%s.%s clashes with %s.%s so it has been renamed %s' % (classname, name, type(f.__self__).__name__, f.__name__, new_name ))
                     name = new_name
                 logging.debug('Adding command : %s -> %s' % (name, value.__name__))
                 self.commands[name] = value
@@ -473,7 +511,8 @@ class Backend(object):
         for name, value in inspect.getmembers(instance_to_inject, inspect.ismethod):
             if getattr(value, '_err_command', False):
                 name = getattr(value, '_err_command_name')
-                del (self.commands[name])
+                if name in self.commands:  # this could happen in premature shutdown
+                    del (self.commands[name])
 
     def warn_admins(self, warning):
         for admin in BOT_ADMINS:
@@ -531,10 +570,10 @@ class Backend(object):
     def send(self, user, text, in_reply_to=None, message_type='chat'):
         """Sends a simple message to the specified user."""
         mess = self.build_message(text)
-        if isinstance(user, basestring):
-            mess.setTo(user)
-        else:
+        if hasattr(user, 'getStripped'):
             mess.setTo(user.getStripped())
+        else:
+            mess.setTo(user)
 
         if in_reply_to:
             mess.setType(in_reply_to.getType())
@@ -574,3 +613,15 @@ class Backend(object):
     @property
     def mode(self):
         raise NotImplementedError("It should be implemented specifically for your backend")
+
+
+def get_jid_from_message(mess):
+    if mess.getType() == 'chat':
+        # strip the resource for direct chats
+        return str(mess.getFrom().getStripped())
+
+    # this is a standard private XMPP reply in MUC
+    fr = mess.getFrom()
+    jid = Identifier(node=fr.node, domain=fr.domain, resource=mess.getMuckNick())
+    logging.debug('Message from MUC. Replying to: %s' % jid)
+    return jid
