@@ -24,27 +24,31 @@ import shutil
 import subprocess
 
 from tarfile import TarFile
-from urllib2 import urlopen
+from urllib.request import urlopen
 
-from errbot import botcmd
-from errbot.backends.base import Backend
+from errbot import botcmd, PY2
+from errbot.backends.base import Backend, HIDE_RESTRICTED_COMMANDS
 
-from errbot.plugin_manager import get_all_active_plugin_names, deactivate_all_plugins, update_plugin_places, get_all_active_plugin_objects,\
-    get_all_plugins, global_restart, get_all_plugin_names, activate_plugin_with_version_check, deactivatePluginByName, get_plugin_obj_by_name,\
-    PluginConfigurationException, check_dependencies
+from errbot.plugin_manager import (
+    get_all_active_plugin_names, deactivate_all_plugins, update_plugin_places, get_all_active_plugin_objects,
+    get_all_plugins, global_restart, get_all_plugin_names, activate_plugin_with_version_check, deactivate_plugin_by_name,
+    get_plugin_obj_by_name, PluginConfigurationException, check_dependencies, reload_plugin_by_name
+)
 
 from errbot.storage import StoreMixin
-from errbot.utils import PLUGINS_SUBDIR, human_name_for_git_url, tail, format_timedelta, which, get_jid_from_message
+from errbot.utils import PLUGINS_SUBDIR, human_name_for_git_url, tail, format_timedelta, which, get_sender_username
 from errbot.repos import KNOWN_PUBLIC_REPOS
 from errbot.version import VERSION
 
-
 def get_class_that_defined_method(meth):
-    for cls in inspect.getmro(meth.im_class):
+    for cls in inspect.getmro(type(meth.__self__)):
         if meth.__name__ in cls.__dict__:
             return cls
     return None
 
+CONFIGS = b'configs' if PY2 else 'configs'
+REPOS = b'repos' if PY2 else 'repos'
+BL_PLUGINS = b'bl_plugins' if PY2 else 'bl_plugins'
 
 class ErrBot(Backend, StoreMixin):
     __errdoc__ = """ Commands related to the bot administration """
@@ -60,22 +64,25 @@ class ErrBot(Backend, StoreMixin):
         self.open_storage(BOT_DATA_DIR + os.sep + 'core.db')
         self.prefix = BOT_PREFIX
         # be sure we have a configs entry for the plugin configurations
-        if 'configs' not in self:
-            self['configs'] = {}
+        if CONFIGS not in self:
+            self[CONFIGS] = {}
         super(ErrBot, self).__init__(*args, **kwargs)
 
     # Repo management
     def get_installed_plugin_repos(self):
-        return self.get('repos', {})
+        return self.get(REPOS, {})
 
     def add_plugin_repo(self, name, url):
+        if PY2:
+            name = name.encode('utf-8')
+            url = url.encode('utf-8')
         repos = self.get_installed_plugin_repos()
         repos[name] = url
-        self['repos'] = repos
+        self[REPOS] = repos
 
     # plugin blacklisting management
     def get_blacklisted_plugin(self):
-        return self.get('bl_plugins', [])
+        return self.get(BL_PLUGINS, [])
 
     def is_plugin_blacklisted(self, name):
         return name in self.get_blacklisted_plugin()
@@ -83,34 +90,36 @@ class ErrBot(Backend, StoreMixin):
     def blacklist_plugin(self, name):
         if self.is_plugin_blacklisted(name):
             logging.warning('Plugin %s is already blacklisted' % name)
-            return
-        self['bl_plugins'] = self.get_blacklisted_plugin() + [name]
+            return 'Plugin %s is already blacklisted' % name
+        self[BL_PLUGINS] = self.get_blacklisted_plugin() + [name]
         logging.info('Plugin %s is now blacklisted' % name)
+        return 'Plugin %s is now blacklisted' % name
 
     def unblacklist_plugin(self, name):
         if not self.is_plugin_blacklisted(name):
             logging.warning('Plugin %s is not blacklisted' % name)
-            return
+            return 'Plugin %s is not blacklisted' % name
         l = self.get_blacklisted_plugin()
         l.remove(name)
-        self['bl_plugins'] = l
-        logging.info('Plugin %s is now unblacklisted' % name)
+        self[BL_PLUGINS] = l
+        logging.info('Plugin %s removed from blacklist' % name)
+        return 'Plugin %s removed from blacklist' % name
 
     # configurations management
     def get_plugin_configuration(self, name):
-        configs = self['configs']
+        configs = self[CONFIGS]
         if name not in configs:
             return None
         return configs[name]
 
     def set_plugin_configuration(self, name, obj):
-        configs = self['configs']
+        configs = self[CONFIGS]
         configs[name] = obj
-        self['configs'] = configs
+        self[CONFIGS] = configs
 
     # this will load the plugins the admin has setup at runtime
     def update_dynamic_plugins(self):
-        all_candidates, errors = update_plugin_places([self.plugin_dir + os.sep + d for d in self.get('repos', {}).keys()])
+        all_candidates, errors = update_plugin_places([self.plugin_dir + os.sep + d for d in self.get(REPOS, {}).keys()])
         self.all_candidates = all_candidates
         return errors
 
@@ -130,13 +139,14 @@ class ErrBot(Backend, StoreMixin):
             for bot in get_all_active_plugin_objects():
                 #noinspection PyBroadException
                 try:
+                    logging.debug('Callback %s' % bot)
                     bot.callback_message(conn, mess)
                 except Exception as _:
                     logging.exception("Crash in a callback_message handler")
 
     def activate_non_started_plugins(self):
         logging.info('Activating all the plugins...')
-        configs = self['configs']
+        configs = self[CONFIGS]
         errors = ''
         for pluginInfo in get_all_plugins():
             try:
@@ -146,7 +156,7 @@ class ErrBot(Backend, StoreMixin):
                 if hasattr(pluginInfo, 'is_activated') and not pluginInfo.is_activated:
                     logging.info('Activate plugin: %s' % pluginInfo.name)
                     activate_plugin_with_version_check(pluginInfo.name, configs.get(pluginInfo.name, None))
-            except Exception, e:
+            except Exception as e:
                 logging.exception("Error loading %s" % pluginInfo.name)
                 errors += 'Error: %s failed to start : %s\n' % (pluginInfo.name, e)
         if errors:
@@ -158,6 +168,7 @@ class ErrBot(Backend, StoreMixin):
             if hasattr(bot, 'callback_connect'):
                 #noinspection PyBroadException
                 try:
+                    logging.debug('Callback %s' % bot)
                     bot.callback_connect()
                 except Exception as _:
                     logging.exception("callback_connect failed for %s" % bot)
@@ -181,6 +192,21 @@ class ErrBot(Backend, StoreMixin):
         self.close_storage()
         logging.info('Bye.')
 
+    @staticmethod
+    def formatted_plugin_list(active_only=True):
+        """
+        Return a formatted, plain-text list of loaded plugins.
+
+        When active_only=True, this will only return plugins which
+        are actually active. Otherwise, it will also include inactive
+        (blacklisted) plugins.
+        """
+        if active_only:
+            all_plugins = get_all_active_plugin_names()
+        else:
+            all_plugins = get_all_plugin_names()
+        return "\n".join(("• " + plugin for plugin in all_plugins))
+
     #noinspection PyUnusedLocal
     @botcmd(template='status')
     def status(self, mess, args):
@@ -192,13 +218,16 @@ class ErrBot(Backend, StoreMixin):
         plugins_statuses = []
         for name in all_attempted:
             if name in all_blacklisted:
-                plugins_statuses.append(('B', name))
+                if name in all_loaded:
+                    plugins_statuses.append(('BL', name))
+                else:
+                    plugins_statuses.append(('BU', name))
             elif name in all_loaded:
                 plugins_statuses.append(('L', name))
             elif get_plugin_obj_by_name(name) is not None and get_plugin_obj_by_name(name).get_configuration_template() is not None and self.get_plugin_configuration(name) is None:
                 plugins_statuses.append(('C', name))
             else:
-                plugins_statuses.append(('E', name))
+                plugins_statuses.append(('U', name))
 
         #noinspection PyBroadException
         try:
@@ -221,14 +250,14 @@ class ErrBot(Backend, StoreMixin):
     def uptime(self, mess, args):
         """ Return the uptime of the bot
         """
-        return 'I up for %s %s (since %s)' % (args, format_timedelta(datetime.now() - self.startup_time), datetime.strftime(self.startup_time, '%A, %b %d at %H:%M'))
+        return "I've been up for %s %s (since %s)" % (args, format_timedelta(datetime.now() - self.startup_time), datetime.strftime(self.startup_time, '%A, %b %d at %H:%M'))
 
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
     def export_configs(self, mess, args):
         """ Returns all the configs in form of a string you can backup
         """
-        return repr(self.get('configs', {}))
+        return repr(self.get(CONFIGS, {}))
 
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
@@ -236,19 +265,19 @@ class ErrBot(Backend, StoreMixin):
         """ Restore the configs from an export from !export configs
         It will merge with preexisting configurations.
         """
-        orig = self.get('configs', {})
+        orig = self.get(CONFIGS, {})
         added = literal_eval(args)
         if type(added) is not dict:
             raise Exception('Weird, it should be a dictionary')
-        self['configs'] = dict(orig.items() + added.items())
-        return "Import is done correctly, there are %i config entries now." % len(self['configs'])
+        self[CONFIGS] = dict(list(orig.items()) + list(added.items()))
+        return "Import is done correctly, there are %i config entries now." % len(self[CONFIGS])
 
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
     def zap_configs(self, mess, args):
         """ WARNING : Deletes all the configuration of all the plugins
         """
-        self['configs'] = {}
+        self[CONFIGS] = {}
         return "Done"
 
     #noinspection PyUnusedLocal
@@ -265,6 +294,7 @@ class ErrBot(Backend, StoreMixin):
         self.send(mess.getFrom(), "Deactivating all the plugins...")
         deactivate_all_plugins()
         self.send(mess.getFrom(), "Restarting")
+        self.shutdown()
         global_restart()
         return "I'm restarting..."
 
@@ -275,42 +305,87 @@ class ErrBot(Backend, StoreMixin):
             if name not in get_all_plugin_names():
                 return "I don't know this %s plugin" % name
             activate_plugin_with_version_check(name, self.get_plugin_configuration(name))
-        except Exception, e:
+        except Exception as e:
             logging.exception("Error loading %s" % name)
             return '%s failed to start : %s\n' % (name, e)
+        get_plugin_obj_by_name(name).callback_connect()
         return "Plugin %s activated" % name
 
     def deactivate_plugin(self, name):
         if name not in get_all_active_plugin_names():
             return "Plugin %s not in active list" % name
-        deactivatePluginByName(name)
+        deactivate_plugin_by_name(name)
         return "Plugin %s deactivated" % name
 
+    #noinspection PyUnusedLocal
+    @botcmd(admin_only=True)
+    def blacklist(self, mess, args):
+        """Blacklist a plugin so that it will not be loaded automatically during bot startup"""
+        if args not in get_all_plugin_names():
+            return ("{} isn't a valid plugin name. The current plugins are:\n"
+                    "{}".format(args, self.formatted_plugin_list(active_only=False)))
+        return self.blacklist_plugin(args)
+
+    #noinspection PyUnusedLocal
+    @botcmd(admin_only=True)
+    def unblacklist(self, mess, args):
+        """Remove a plugin from the blacklist"""
+        if args not in get_all_plugin_names():
+            return ("{} isn't a valid plugin name. The current plugins are:\n"
+                    "{}".format(args, self.formatted_plugin_list(active_only=False)))
+        return self.unblacklist_plugin(args)
+
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
     def load(self, mess, args):
         """load a plugin"""
-        self.unblacklist_plugin(args)
-        return self.activate_plugin(args)
+        args = args.strip()
+        if not args:
+            return ("Please tell me which of the following plugins to reload:\n"
+                    "{}".format(self.formatted_plugin_list(active_only=False)))
+        if args not in get_all_plugin_names():
+            return ("{} isn't a valid plugin name. The current plugins are:\n"
+                    "{}".format(args, self.formatted_plugin_list(active_only=False)))
+        if args in get_all_active_plugin_names():
+            return "{} is already loaded".format(args)
+
+        reload_plugin_by_name(args)
+        r = self.activate_plugin(args)
+        return r
 
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
     def unload(self, mess, args):
         """unload a plugin"""
+        args = args.strip()
+        if not args:
+            return ("Please tell me which of the following plugins to reload:\n"
+                    "{}".format(self.formatted_plugin_list(active_only=False)))
+        if args not in get_all_plugin_names():
+            return ("{} isn't a valid plugin name. The current plugins are:\n"
+                    "{}".format(args, self.formatted_plugin_list(active_only=False)))
         if args not in get_all_active_plugin_names():
-            return '%s in not active' % args
-        self.blacklist_plugin(args)
+            return "{} is not currently loaded".format(args)
+
         return self.deactivate_plugin(args)
 
     #noinspection PyUnusedLocal
     @botcmd(admin_only=True)
     def reload(self, mess, args):
         """reload a plugin"""
-        if self.is_plugin_blacklisted(args):
-            self.unblacklist_plugin(args)
-        result = "%s / %s" % (self.deactivate_plugin(args), self.activate_plugin(args))
-        get_plugin_obj_by_name(args).callback_connect()
-        return result
+        args = args.strip()
+        if not args:
+            yield ("Please tell me which of the following plugins to reload:\n"
+                    "{}".format(self.formatted_plugin_list(active_only=False)))
+            return
+        if args not in get_all_plugin_names():
+            yield ("{} isn't a valid plugin name. The current plugins are:\n"
+                    "{}".format(args, self.formatted_plugin_list(active_only=False)))
+            return
+
+        yield self.deactivate_plugin(args)  # Not needed but keeps the feedback to user consistent
+        reload_plugin_by_name(args)
+        yield self.activate_plugin(args)
 
     @botcmd(admin_only=True)
     def repos_install(self, mess, args):
@@ -335,8 +410,8 @@ class ErrBot(Backend, StoreMixin):
         else:
             human_name = human_name_for_git_url(args)
             p = subprocess.Popen([git_path, 'clone', args, human_name], cwd=self.plugin_dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            feedback = p.stdout.read()
-            error_feedback = p.stderr.read()
+            feedback = p.stdout.read().decode('utf-8')
+            error_feedback = p.stderr.read().decode('utf-8')
             if p.wait():
                 return "Could not load this plugin : \n%s\n---\n%s" % (feedback, error_feedback)
         self.add_plugin_repo(human_name, args)
@@ -354,7 +429,7 @@ class ErrBot(Backend, StoreMixin):
         """
         if not args.strip():
             return "You should have a repo name as argument"
-        repos = self.get('repos', {})
+        repos = self.get(REPOS, {})
         if args not in repos:
             return "This repo is not installed check with " + self.prefix + "repos the list of installed ones"
 
@@ -366,7 +441,7 @@ class ErrBot(Backend, StoreMixin):
 
         shutil.rmtree(plugin_path)
         repos.pop(args)
-        self['repos'] = repos
+        self[REPOS] = repos
 
         return 'Plugins unloaded and repo %s removed' % args
 
@@ -379,9 +454,9 @@ class ErrBot(Backend, StoreMixin):
         all_names = sorted(set([name for name in KNOWN_PUBLIC_REPOS] + [name for name in installed_repos]))
         max_width = max([len(name) for name in all_names])
         return {'repos': [
-            (repo_name in installed_repos, repo_name in KNOWN_PUBLIC_REPOS, repo_name.ljust(max_width), KNOWN_PUBLIC_REPOS[repo_name][1]
-                if repo_name in KNOWN_PUBLIC_REPOS else installed_repos[repo_name])
-            for repo_name in all_names]}
+        (repo_name in installed_repos, repo_name in KNOWN_PUBLIC_REPOS, repo_name.ljust(max_width), KNOWN_PUBLIC_REPOS[repo_name][1]
+        if repo_name in KNOWN_PUBLIC_REPOS else installed_repos[repo_name])
+        for repo_name in all_names]}
 
     def get_doc(self, command):
         """Get command documentation
@@ -403,34 +478,40 @@ class ErrBot(Backend, StoreMixin):
         usage = ''
         if not args:
             description = 'Available help:\n'
-            command_classes = sorted(set(self.get_command_classes()))
+            command_classes = sorted(set(self.get_command_classes()), key = lambda c:c.__name__)
             usage = '\n'.join(self.prefix + 'help %s: %s' % (clazz.__name__, clazz.__errdoc__ or '(undocumented)') for clazz in command_classes)
         elif args == 'full':
             description = 'Available commands:'
 
             clazz_commands = {}
-            for (name, command) in self.commands.iteritems():
+            for (name, command) in self.commands.items():
                 clazz = get_class_that_defined_method(command)
                 commands = clazz_commands.get(clazz, [])
-                commands.append((name, command))
-                clazz_commands[clazz] = commands
+                if not HIDE_RESTRICTED_COMMANDS or self.check_command_access(mess, name)[0]:
+                    commands.append((name, command))
+                    clazz_commands[clazz] = commands
 
-            for clazz in sorted(clazz_commands):
+            for clazz in sorted(set(clazz_commands), key = lambda c:c.__name__):
                 usage += '\n\n%s: %s\n' % (clazz.__name__, clazz.__errdoc__ or '')
                 usage += '\n'.join(sorted([
-                    '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1),
-                    (self.get_doc(command).strip()).split('\n', 1)[0])
-                    for (name, command) in clazz_commands[clazz] if name != 'help' and not command._err_command_hidden
+                '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1),
+                                                 (self.get_doc(command).strip()).split('\n', 1)[0])
+                for (name, command) in clazz_commands[clazz] 
+                    if name != 'help' 
+                    and not command._err_command_hidden 
+                    and (not HIDE_RESTRICTED_COMMANDS or self.check_command_access(mess, name)[0])
                 ]))
             usage += '\n\n'
         elif args in (clazz.__name__ for clazz in self.get_command_classes()):
             # filter out the commands related to this class
-            commands = [(name, command) for (name, command) in self.commands.iteritems() if get_class_that_defined_method(command).__name__ == args]
+            commands = [(name, command) for (name, command) in self.commands.items() if get_class_that_defined_method(command).__name__ == args]
             description = 'Available commands for %s:\n\n' % args
             usage += '\n'.join(sorted([
-                '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1),
-                (self.get_doc(command).strip()).split('\n', 1)[0])
-                for (name, command) in commands if not command._err_command_hidden
+            '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1),
+                                             (self.get_doc(command).strip()).split('\n', 1)[0])
+            for (name, command) in commands 
+                if not command._err_command_hidden  
+                and (not HIDE_RESTRICTED_COMMANDS or self.check_command_access(mess, name)[0])
             ]))
         else:
             return super(ErrBot, self).help(mess, '_'.join(args.strip().split(' ')))
@@ -444,9 +525,10 @@ class ErrBot(Backend, StoreMixin):
     def history(self, mess, args):
         """display the command history"""
         answer = []
-        l = len(self.cmd_history)
+        user_cmd_history = self.cmd_history[get_sender_username(mess)]
+        l = len(user_cmd_history)
         for i in range(0, l):
-            c = self.cmd_history[i]
+            c = user_cmd_history[i]
             answer.append('%2i:%s%s %s' % (l - i, self.prefix, c[0], c[1]))
         return '\n'.join(answer)
 
@@ -471,18 +553,20 @@ class ErrBot(Backend, StoreMixin):
         description = 'Available commands:\n'
 
         clazz_commands = {}
-        for (name, command) in self.commands.iteritems():
+        for (name, command) in self.commands.items():
             clazz = get_class_that_defined_method(command)
+            clazz = str.__module__ + '.' + clazz.__name__  # makes the fuul qualified name
             commands = clazz_commands.get(clazz, [])
-            commands.append((name, command))
-            clazz_commands[clazz] = commands
+            if not HIDE_RESTRICTED_COMMANDS or self.check_command_access(mess, name)[0]:
+                commands.append((name, command))
+                clazz_commands[clazz] = commands
 
         usage = ''
         for clazz in sorted(clazz_commands):
             usage += '\n'.join(sorted([
-                '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1), (command.__doc__ or '(undocumented)').strip().split('\n', 1)[0])
-                for (name, command) in clazz_commands[clazz] if
-                args is not None and command.__doc__ is not None and args.lower() in command.__doc__.lower().decode('utf-8', 'ignore') and name != 'help' and not command._err_command_hidden
+            '\t' + self.prefix + '%s: %s' % (name.replace('_', ' ', 1), (command.__doc__ or '(undocumented)').strip().split('\n', 1)[0])
+            for (name, command) in clazz_commands[clazz] if
+            args is not None and command.__doc__ is not None and args.lower() in command.__doc__.lower() and name != 'help' and not command._err_command_hidden
             ]))
         usage += '\n\n'
 
@@ -505,7 +589,7 @@ class ErrBot(Backend, StoreMixin):
             return 'git command not found: You need to have git installed on your system to by able to update git based plugins.'
 
         directories = set()
-        repos = self.get('repos', {})
+        repos = self.get(REPOS, {})
         core_to_update = 'all' in args or 'core' in args
         if core_to_update:
             directories.add(os.path.dirname(__file__))
@@ -518,8 +602,8 @@ class ErrBot(Backend, StoreMixin):
         for d in directories:
             self.send(mess.getFrom(), "I am updating %s ..." % d, message_type=mess.getType())
             p = subprocess.Popen([git_path, 'pull'], cwd=d, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            feedback = p.stdout.read() + '\n' + '-' * 50 + '\n'
-            err = p.stderr.read().strip()
+            feedback = p.stdout.read().decode('utf-8') + '\n' + '-' * 50 + '\n'
+            err = p.stderr.read().strip().decode('utf-8')
             if err:
                 feedback += err + '\n' + '-' * 50 + '\n'
             dep_err = check_dependencies(d)
@@ -534,13 +618,9 @@ class ErrBot(Backend, StoreMixin):
                         if plugin.path.startswith(d) and hasattr(plugin, 'is_activated') and plugin.is_activated:
                             name = plugin.name
                             self.send(mess.getFrom(), '/me is reloading plugin %s' % name)
-                            self.deactivate_plugin(plugin.name)                 # calm the plugin down
-                            module = __import__(plugin.path.split(os.sep)[-1])  # find back the main module of the plugin
-                            reload(module)                                      # reload it
-                            class_name = type(plugin.plugin_object).__name__    # find the original name of the class
-                            newclass = getattr(module, class_name)              # retreive the corresponding new class
-                            plugin.plugin_object.__class__ = newclass           # BAM, declare the instance of the new type
-                            self.activate_plugin(plugin.name)                   # wake the plugin up
+                            reload_plugin_by_name(plugin.name)
+                            self.activate_plugin(plugin.name)
+                            self.send(mess.getFrom(), '%s reloaded and reactivated' % name)
         if core_to_update:
             self.restart(mess, '')
             return "You have updated the core, I need to restart."
@@ -592,7 +672,7 @@ class ErrBot(Backend, StoreMixin):
         self.deactivate_plugin(plugin_name)
         try:
             self.activate_plugin(plugin_name)
-        except PluginConfigurationException, ce:
+        except PluginConfigurationException as ce:
             logging.debug('Invalid configuration for the plugin, reverting the plugin to unconfigured')
             self.set_plugin_configuration(plugin_name, None)
             return 'Incorrect plugin configuration: %s' % ce
@@ -613,4 +693,4 @@ class ErrBot(Backend, StoreMixin):
         if BOT_LOG_FILE:
             with open(BOT_LOG_FILE, 'r') as f:
                 return tail(f, n)
-        return 'No log is configured, please define BOT_LOG_FILE in config.py'
\ No newline at end of file
+        return 'No log is configured, please define BOT_LOG_FILE in config.py'
