@@ -1,5 +1,3 @@
 # This class just hold the global singletons for the bot outside of the plugin framework to avoid spurrious modules reloads
-from flask.app import Flask
-
+plugin_manager = None
 bot = None
-flask_app = Flask(__name__)
\ No newline at end of file
