@@ -1,13 +1,14 @@
 import logging
 import os
 from jinja2 import Environment, FileSystemLoader
-
+from bottle import TEMPLATE_PATH
 
 def make_templates_path(root):
     return root + os.sep + 'templates'
 
-
-template_path = [make_templates_path(os.path.dirname(__file__))]  # the system templates
+system_templates_path = make_templates_path(os.path.dirname(__file__))
+template_path = [system_templates_path]
+TEMPLATE_PATH.insert(0, system_templates_path)  # for views
 env = Environment(loader=FileSystemLoader(template_path))
 
 
@@ -24,7 +25,8 @@ def add_plugin_templates_path(path):
     tmpl_path = make_templates_from_plugin_path(path)
     if os.path.exists(tmpl_path):
         logging.debug("Templates directory found for this plugin [%s]" % tmpl_path)
-        template_path.append(tmpl_path)
+        template_path.append(tmpl_path)  # for webhooks
+        TEMPLATE_PATH.insert(0, tmpl_path)  # for webviews
         env = Environment(loader=FileSystemLoader(template_path))  # ditch and recreate a new templating environment
         return
     logging.debug("No templates directory found for this plugin [Looking for %s]" % tmpl_path)
@@ -34,5 +36,6 @@ def remove_plugin_templates_path(path):
     global env
     tmpl_path = make_templates_from_plugin_path(path)
     if tmpl_path in template_path:
-        template_path.pop(template_path.index(tmpl_path))
+        template_path.remove(tmpl_path)  # for webhooks
+        TEMPLATE_PATH.remove(tmpl_path)  # for webviews
         env = Environment(loader=FileSystemLoader(template_path))  # ditch and recreate a new templating environment
\ No newline at end of file
